from io import IOBase
from typing import IO, Union
from .base import BaseCompresser, PathType, PATH_TYPES
from .registry import register_compresser


class Lz4Compresser(BaseCompresser):
    """Compresser class that wraps the lz4 compression package.

    This class relies on the :mod:`lz4` module to open the input/output binary stream where the
    pickled python objects will be written to (or read from). During an instance's initialization,
    the binary stream is opened using ``lz4.frame.open(path, mode=mode, **kwargs)``.

    Parameters
    ----------
    path : Union[PathType, IO[bytes]]
        A PathType object (``str``, ``bytes``, ``os.PathType``) or a file-like
        object (e.g. ``io.BaseIO`` instances). The path that will be used to open the
        input/output binary stream.
    mode : str
        Mode with which to open the file buffer.
    kwargs
        Any other key word arguments that are passed to :func:`lz4.frame.open`.
    """

    def __init__(self, path: Union[PathType, IO[bytes]], mode: str, **kwargs):
        try:
            import lz4.frame
        except ImportError:
            raise RuntimeError(
                "The lz4 compression protocol requires the lz4 package to be installed. "
                "Please pip install lz4 and retry."
            )
        if not isinstance(path, PATH_TYPES + (IOBase,)):
            raise TypeError(f"Unhandled path type {type(path)}")
        self._stream = lz4.frame.open(path, mode=mode, **kwargs)

    def close(self):
        self._stream.close()

    def get_stream(self) -> IO[bytes]:
        return self._stream  # type: ignore


register_compresser(
    compression="lz4",
    compresser=Lz4Compresser,
    extensions=["lz4"],
    default_write_mode="wb",
    default_read_mode="rb",
)



from typing import Union, Optional, IO
import codecs
from os import PathLike
from os.path import splitext
from .compressers import (
    BaseCompresser,
    get_compresser,
    get_compression_from_extension,
    get_compression_read_mode,
    get_compression_write_mode,
    get_default_compression_mapping,
)


PATH_TYPES = (str, bytes, PathLike)
PathType = Union[str, bytes, PathLike]
FileType = IO[bytes]


__all__ = [
    "instantiate_compresser",
]


def _stringyfy_path(path: PathType) -> str:
    """Convert a path that is a PATH_TYPES instance to a string.

    Parameters
    ----------
    path : PathType
        If ``path`` is a ``string`` instance, it is returned as is.
        If ``path`` is a ``bytes`` instance, it is decoded with utf8 codec.
        If it is ``os.PathLike`` or ``pathlib.PurePath`` then it is converted
        to a string with ``str(path)``.

    Returns
    -------
    path_string : str
        The string representation of the ``path``.

    Raises
    ------
    TypeError
        If the supplied ``path`` is not a ``PATH_TYPES`` instance.

    """
    if not isinstance(path, PATH_TYPES):
        raise TypeError(
            "Cannot convert supplied path type to string. Supplied path: {}, "
            "type: {}".format(path, type(path))
        )
    if isinstance(path, bytes):
        path = codecs.decode(path, "utf-8")
    else:
        path = str(path)
    return path


def instantiate_compresser(
    compression: Optional[str],
    path: Union[PathType, FileType],
    mode: str,
    set_default_extension: bool = True,
    **kwargs,
) -> BaseCompresser:
    """Initialize the compresser instance for the supplied path and compression.

    Initiate a :class:`compress_pickle.compressers.base.BaseCompresser` instance based on the
    supplied (or inferred) ``compression`` method. This instance will open a file like object
    with supplied ``mode``, using the provided ``path``. Furthermore, when the ``path`` is not
    a binary stream, this function can also potentially set the path's extension to the default
    extension registered to the used compression method.

    Parameters
    ----------
    compression : Optional[str]
        The compression method name. Refer to
        :func:`~compress_pickle.compressers.registry.get_known_compressions` for a list of the
        known compression methods. If ``"infer"``, the compression method is inferred from the
        ``path`` extension (only possible if ``path`` is a ``PathType``). Refer to
        :func:`~compress_pickle.compressers.registry.get_registered_extensions` for the mapping
        between extensions and compression methods.
    path : Union[PathType, FileType]
        A path-like object (``str``, ``bytes``, ``os.PathType``) or a file-like
        object (``io.BaseIO`` instances) that will be passed to the
        :class:`compress_pickle.compressers.base.BaseCompresser` class.
    mode : str
        Mode with which to open the file-like stream. Only used if the ``path`` is a
        ``PathType``. If "read", the default read mode is automatically assigned by
        :func:`~compress_pickle.compressers.registry.get_compression_read_mode` based on the used
        compression method. If "write", the default write mode is automatically assigned from
        :func:`~compress_pickle.compressers.registry.get_compression_write_mode` based on the used
        compression method.
    set_default_extension : bool
        If ``True``, the default extension given the provided compression protocol is set to the
        supplied ``path``. Refer to
        :func:`~compress_pickle.compressers.registry.get_default_compression_mapping` for the
        default extension registered to each compression method.
    **kwargs
        Any extra keyword arguments are passed to the
        :class:`compress_pickle.compressers.base.BaseCompresser` initialization.

    Returns
    -------
    BaseCompresser
        The compresser instance that will be used to create the byte stream from which a
        :class:`compress_pickle.picklers.base.BasePicklerIO` will read or write serialized objects.
    """
    if isinstance(path, PATH_TYPES):
        _path = _stringyfy_path(path)
    if compression == "infer":
        compression = _infer_compression_from_path(_path)
    compresser_class = get_compresser(compression)
    if set_default_extension and isinstance(path, PATH_TYPES):
        _path = _set_default_extension(_path, compression)
    if mode == "write":
        mode = get_compression_write_mode(compression)
    elif mode == "read":
        mode = get_compression_read_mode(compression)
    compresser = compresser_class(
        _path if isinstance(path, PATH_TYPES) else path, mode=mode, **kwargs
    )
    return compresser


def _infer_compression_from_path(path: PathType) -> Optional[str]:
    if not isinstance(path, PATH_TYPES):
        raise TypeError(
            f"Cannot infer the compression from a path that is not an instance of "
            f"{PATH_TYPES}. Encountered {type(path)}"
        )
    _, extension = splitext(_stringyfy_path(path))
    return get_compression_from_extension(extension)


def _set_default_extension(path: PathType, compression: Optional[str]) -> str:
    root, current_ext = splitext(_stringyfy_path(path))
    return root + "." + get_default_compression_mapping()[compression]



from . import base
from . import registry
from . import no_compression
from . import gzip
from . import bz2
from . import lzma
from . import zipfile
from . import lz4
from .base import BaseCompresser
from .registry import *
from .no_compression import NoCompresser
from .gzip import GzipCompresser
from .bz2 import Bz2Compresser
from .lzma import LzmaCompresser
from .zipfile import ZipfileCompresser
from .lz4 import Lz4Compresser



from typing import Any, IO
import marshal
from .base import BasePicklerIO
from .registry import register_pickler


__all__ = ["MarshalPicklerIO"]


class MarshalPicklerIO(BasePicklerIO):
    """A PicklerIO class that wraps standard :func:`marshal.dump` and :func:`marshal.load`."""

    def dump(self, obj: Any, stream: IO[bytes], **kwargs):
        """Write a serialized binary representation of an object to a stream.

        Parameters
        ----------
        obj : Any
            The python object that must be serialized and written.
        stream : IO[bytes]
            The binary stream (file-like object) where the serialized object must be written to.
        kwargs
            Any extra keyword arguments to pass to :func:`marshal.dump`.
        """
        args = kwargs.pop("version", tuple())
        args += tuple(kwargs.values())
        marshal.dump(obj, stream, *args)

    def load(self, stream: IO[bytes], **kwargs):
        """Load a serialized binary representation of an object from a stream.

        Parameters
        ----------
        stream : IO[bytes]
            The binary stream (file-like object) from where the serialized object must be loaded.
        kwargs
            Any extra keyword arguments to pass to :func:`marshal.load`.

        Returns
        -------
        obj : Any
            The python object that was loaded.
        """
        return marshal.load(stream)


register_pickler("marshal", MarshalPicklerIO)



from typing import Union, IO
from abc import abstractmethod
from os import PathLike


PATH_TYPES = (str, bytes, PathLike)
PathType = Union[str, bytes, PathLike]


class BaseCompresser:
    """Compresser abstract base class.

    This class is in charge of handing the binary stream where the pickled python objects must
    be written to (or read from). During an instance's initialization, the binary stream must
    be opened based on a supplied ``path`` parameter and ``mode``. This stream is nothing more than
    a python "file-like" object that is in charge of actually writting and reading the binary
    contents.

    Parameters
    ----------
    path : Union[PathType, IO[bytes]]
        A PathType object (``str``, ``bytes``, ``os.PathType``) or a file-like
        object (e.g. ``io.BaseIO`` instances). The path that will be used to open the
        input/output binary stream.
    mode : str
        Mode with which to open the file buffer.
    kwargs
        Any other key word arguments that can be handled by the specific binary stream opener.
    """

    @abstractmethod
    def __init__(
        self,
        path: Union[PathType, IO[bytes]],
        mode: str,
        **kwargs,
    ):  # pragma: no cover
        pass

    @abstractmethod
    def close(self, **kwargs):  # pragma: no cover
        """Close the input/output binary stream.

        This closes the input/output stream that is created during the ``__init__``.

        Parameters
        ----------
        kwargs
            Any other key word arguments that can be handled by the specific binary stream closer.
        """
        pass

    @abstractmethod
    def get_stream(self) -> IO[bytes]:  # pragma: no cover
        """Get the input/output binary stream (i.e. a file-like object).

        Returns
        -------
        IO[bytes]
            The input/output binary stream where the pickled objects are written to or read from.
        """
        pass



from io import IOBase
from typing import IO, Union
from .base import BaseCompresser, PathType, PATH_TYPES
from .registry import register_compresser, add_compression_alias


class NoCompresser(BaseCompresser):
    """Compresser class that represents a simple uncompressed file object.

    This class either simply calls ``open`` on the supplied ``path`` or uses the ``path`` as
    the binary input/output stream.

    Parameters
    ----------
    path : Union[PathType, IO[bytes]]
        A PathType object (``str``, ``bytes``, ``os.PathType``) or a file-like
        object (e.g. ``io.BaseIO`` instances). The path that will be used to open the
        input/output binary stream.
    mode : str
        Mode with which to open the file buffer.
    kwargs
        Any other key word arguments that are passed to :func:`open`.
    """

    def __init__(self, path: Union[PathType, IO[bytes]], mode: str, **kwargs):
        if isinstance(path, PATH_TYPES):
            self._must_close = True
            self._stream = open(file=path, mode=mode, **kwargs)
        elif isinstance(path, IOBase):
            self._must_close = False
            self._stream = path
        else:
            raise TypeError(f"Unhandled path type {type(path)}")

    def close(self):
        """Close the input/output stream if necessary.

        This will only close the stream if the ``path`` argument passed during construction was a
        ``PathType`` (``str``, ``bytes``, ``os.PathType``). Otherwise, this method doesn't do
        anything.
        """
        if self._must_close:
            self._stream.close()

    def get_stream(self) -> IO[bytes]:
        return self._stream


register_compresser(
    compression=None,
    compresser=NoCompresser,
    extensions=["pkl", "pickle"],
    default_write_mode="wb",
    default_read_mode="rb",
)
add_compression_alias("pickle", None)



from .base import BasePicklerIO
from .registry import *
from .pickle import BuiltinPicklerIO
from .optimized_pickle import OptimizedPicklerIO
from .marshal import MarshalPicklerIO
from .dill import DillPicklerIO
from .cloudpickle import CloudPicklerIO



from os.path import basename
from io import IOBase
from typing import IO, Union
import zipfile
from .base import BaseCompresser, PathType, PATH_TYPES
from .registry import register_compresser


class ZipfileCompresser(BaseCompresser):
    """Compresser class that wraps the zipfile compression package.

    This class relies on the :mod:`zipfile` module to open the input/output binary stream where the
    pickled python objects will be written to (or read from).
    During an instance's initialization, a :class:`zipfile.ZipFile` instance is created around
    the supplied ``path``. The opened ``ZipFile`` is called the archive and works as a kind
    directory, that can hold other directories or files. These are called members of the archive.
    The ``ZipfileCompresser`` creates the input/output stream by opening a member file in the
    opened ``ZipFile`` archive. The name of the archive member can be chosen with the ``arcname``
    argument.

    Parameters
    ----------
    path : Union[PathType, IO[bytes]]
        A PathType object (``str``, ``bytes``, ``os.PathType``) or a file-like
        object (e.g. ``io.BaseIO`` instances). The path that will be used to open the
        ``zipfile.Zipfile`` object. The input/output binary stream is then opened from the
        aforementioned object as ``zipfile.Zipfile(*).open(arcname, mode)``.
    mode : str
        Mode with which to open the ``Zipfile`` (and also the archive member that is used as the
        input/output binary stream).
    arcname : Optional[str]
        The name of the archive member of the opened ``zipfile.Zipfile`` that will be used as
        the binary input/output stream.
        If ``None``, the ``arcname`` is assumed to be the basename of ``path`` (when ``path``
        is path-like), ``path.name`` (when ``path`` is file-like and it has a name attribute) or
        "default" otherwise.
    pwd : Optional[str]
        The password used to decrypt encrypted ZIP files.
    zipfile_compression : Optional[str]
        If not ``None``, it is passed as the ``compression`` keyword argument to
        ``zipfile.Zipfile(...)``.
    kwargs
        Any other key word arguments that are passed to ``zipfile.ZipFile``.
    """

    def __init__(
        self,
        path: Union[PathType, IO[bytes]],
        mode: str,
        *,
        arcname=None,
        pwd=None,
        zipfile_compression=None,
        **kwargs,
    ):
        if zipfile_compression is not None:
            kwargs["compression"] = zipfile_compression
        if not isinstance(path, PATH_TYPES + (IOBase,)):
            raise TypeError(f"Unhandled path type {type(path)}")
        self._arch = zipfile.ZipFile(path, mode=mode, **kwargs)  # type: ignore
        if arcname is None:
            if isinstance(path, PATH_TYPES):
                file_path = basename(path)
            else:
                file_path = getattr(path, "name", "default")
            arcname = file_path
        else:
            file_path = arcname
        self._stream = self._arch.open(file_path, mode=mode, pwd=pwd)  # type: ignore

    def close(self):
        """Close the input/output binary stream and the ``ZipFile``.

        This closes the ``zipfile.ZipFile`` instance and archive member file-objects that
        are created during the ``__init__``.
        """
        self._stream.close()
        self._arch.close()

    def get_stream(self) -> IO[bytes]:
        return self._stream


register_compresser(
    compression="zipfile",
    compresser=ZipfileCompresser,
    extensions=["zip"],
    default_write_mode="w",
    default_read_mode="r",
)



from typing import Dict, List, Type
from .base import BasePicklerIO


__all__ = [
    "get_pickler",
    "register_pickler",
    "get_known_picklers",
    "list_registered_picklers",
]


class _pickler_registry:
    _pickler_registry: Dict[str, Type[BasePicklerIO]] = {}
    _pickler_aliases: Dict[str, str] = {}

    @classmethod
    def get_pickler(cls, name: str) -> Type[BasePicklerIO]:
        """Get the PicklerIO class registered with a given pickler name.

        Parameters
        ----------
        name: str
            The pickler name.

        Raises
        ------
        ValueError
            If the supplied ``name`` has not been registered.

        Returns
        -------
        BasePicklerIO
            The PicklerIO class associated to the pickler ``name``.
        """
        try:
            return cls._pickler_registry[name]
        except Exception:
            raise ValueError(
                f"Unknown pickler {name}. "
                "Available values are {list(cls._pickler_registry)}"
            )

    @classmethod
    def register_pickler(
        cls,
        name: str,
        pickler: Type[BasePicklerIO],
    ):
        """Register a pickler handler.

        Parameters
        ----------
        name: str
            The pickler name that will be registered.
        pickler : Type[BasePicklerIO]
            The PicklerIO class. This should be a :class:`~compress_pickle.picklers.base.BasePicklerIO`
            subclass.

        Raises
        ------
        ValueError
            If the supplied ``name`` is already contained in the registry.
        TypeError
            If the supplied pickler is not a :class:`~compress_pickle.picklers.base.BasePicklerIO`
            subclass.
        """
        if name in cls._pickler_registry:
            raise ValueError(
                f"A pickler with name {name} is already registered. "
                "Please choose a different name."
            )
        try:
            _subclass = issubclass(pickler, BasePicklerIO)
            if not _subclass:
                raise TypeError()
        except Exception:
            raise TypeError(
                f"The supplied pickler {pickler} is not a derived from {BasePicklerIO}"
            )
        cls._pickler_registry[name] = pickler

    @classmethod
    def add_pickler_alias(cls, alias: str, pickler: str):
        """Add an alias for an already registered pickler.

        Parameters
        ----------
        alias : str
            The alias to register
        compression : str
            The pickler name that must already be registered.

        Raises
        ------
        ValueError
            If the supplied ``pickler`` name is not known or if the supplied ``alias``
            is already contained in the registry.
        """
        if alias in cls._pickler_registry:
            raise ValueError(
                f"The alias {alias} is already registered, please choose a different alias."
            )
        if pickler not in cls._pickler_registry:
            raise ValueError(
                "Unknown pickler name {}. Available values are: {}".format(
                    pickler, list(cls._pickler_registry)
                )
            )
        cls._pickler_registry[alias] = cls._pickler_registry[pickler]
        cls._pickler_aliases[alias] = pickler


get_pickler = _pickler_registry.get_pickler

register_pickler = _pickler_registry.register_pickler

add_pickler_alias = _pickler_registry.add_pickler_alias


def get_known_picklers() -> List[str]:
    """Get a list of known pickle serialization methods.

    Returns
    -------
    List[str]
        List of known pickler serialization methods.
    """
    return list(_pickler_registry._pickler_registry)


def list_registered_picklers() -> List[Type[BasePicklerIO]]:  # pragma: no cover
    """Get the list of registered PicklerIO classes.

    Returns
    -------
    List[BasePicklerIO]
        The list of registered pickler classes.
    """
    return list(_pickler_registry._pickler_registry.values())



from typing import Any, IO
import pickle
from .base import BasePicklerIO
from .registry import register_pickler


__all__ = ["BuiltinPicklerIO"]


class BuiltinPicklerIO(BasePicklerIO):
    """A PicklerIO class that wraps standard :func:`pickle.dump` and :func:`pickle.load`."""

    def dump(self, obj: Any, stream: IO[bytes], **kwargs):
        """Write a serialized binary representation of an object to a stream.

        Parameters
        ----------
        obj : Any
            The python object that must be serialized and written.
        stream : IO[bytes]
            The binary stream (file-like object) where the serialized object must be written to.
        kwargs
            Any extra keyword arguments to pass to :func:`pickle.dump`.
        """
        if kwargs.get("protocol", None) == 5:
            stream.write(pickle.dumps(obj, **kwargs))
        else:
            pickle.dump(obj, stream, **kwargs)

    def load(self, stream: IO[bytes], **kwargs):
        """Load a serialized binary representation of an object from a stream.

        Parameters
        ----------
        stream : IO[bytes]
            The binary stream (file-like object) from where the serialized object must be loaded.
        kwargs
            Any extra keyword arguments to pass to :func:`pickle.load`.

        Returns
        -------
        obj : Any
            The python object that was loaded.
        """
        return pickle.load(stream, **kwargs)


register_pickler("pickle", BuiltinPicklerIO)



from typing import Any, IO
from .base import BasePicklerIO
from .registry import register_pickler

try:
    import cloudpickle

    _cloudpickle_available = True
except ImportError:  # pragma: no cover
    _cloudpickle_available = False


__all__ = ["CloudPicklerIO"]


class CloudPicklerIO(BasePicklerIO):
    """A PicklerIO class that wraps ``cloudpickle.dump`` and ``cloudpickle.load``."""

    def __init__(self):
        if not _cloudpickle_available:
            raise RuntimeError(
                "The cloudpickle serialization protocol requires the cloudpickle package to be "
                "installed. Please pip install cloudpickle and retry."
            )
        super().__init__()

    def dump(self, obj: Any, stream: IO[bytes], **kwargs):
        """Write a serialized binary representation of an object to a stream.

        Parameters
        ----------
        obj : Any
            The python object that must be serialized and written.
        stream : IO[bytes]
            The binary stream (file-like object) where the serialized object must be written to.
        kwargs
            Any extra keyword arguments to pass to ``cloudpickle.dump``.
        """
        cloudpickle.dump(obj, stream, **kwargs)

    def load(self, stream: IO[bytes], **kwargs):
        """Load a serialized binary representation of an object from a stream.

        Parameters
        ----------
        stream : IO[bytes]
            The binary stream (file-like object) from where the serialized object must be loaded.
        kwargs
            Any extra keyword arguments to pass to ``cloudpickle.load``.

        Returns
        -------
        obj : Any
            The python object that was loaded.
        """
        return cloudpickle.load(stream, **kwargs)


register_pickler("cloudpickle", CloudPicklerIO)
